import json
from typing import Iterable

import numpy as np
import pandas as pd

from starfish.types import Axes, Features
from ._validated_table import ValidatedTable


class SpotAttributes(ValidatedTable):

    required_fields = {
        Axes.X.value,          # spot x-coordinate
        Axes.Y.value,          # spot y-coordinate
        Axes.ZPLANE.value,     # spot z-coordinate
        Features.SPOT_RADIUS,  # spot radius
    }

    def __init__(self, spot_attributes: pd.DataFrame) -> None:
        """Construct a SpotAttributes instance

        Parameters
        ----------
        spot_attributes : pd.DataFrame

        """
        super().__init__(spot_attributes, SpotAttributes.required_fields)

    @classmethod
    def empty(cls, extra_fields: Iterable = tuple()) -> "SpotAttributes":
        """return an empty SpotAttributes object"""
        fields = list(cls.required_fields.union(extra_fields))
        dtype = list(zip(fields, [np.object] * len(fields)))
        return cls(pd.DataFrame(np.array([], dtype=dtype)))

    def save_geojson(self, output_file_name: str) -> None:
        """Save to geojson for web visualization

        Parameters
        ----------
        output_file_name : str
            name of output json file

        """

        # TODO ambrosejcarr: write a test for this
        geojson = [
            {
                'properties': {'id': int(row.spot_id), 'radius': int(row.r)},
                'geometry': {'type': 'Point', 'coordinates': [int(row.x), int(row.y)]}
            } for index, row in self.data.iterrows()
        ]

        with open(output_file_name, 'w') as f:
            f.write(json.dumps(geojson))



import subprocess
import unittest

from starfish.util import exec


def assert_contains(actual, expected):
    if isinstance(actual, bytes):
        actual = actual.decode("utf-8")
    if expected not in actual:
        raise Exception(f"counldn't find: 'f{expected}'")


class TestHelpReturnCode(unittest.TestCase):
    """
    Tests that the CLI supports a '--help' option at all of the expected levels.
    """

    STAGES = (
        [
            "starfish", "--help",
            lambda tempdir: tempdir
        ],
        [
            "starfish", "detect_spots", "--help",
            lambda tempdir: tempdir
        ],
        [
            "starfish", "detect_spots", "BlobDetector", "--help",
            lambda tempdir: tempdir
        ],
    )

    def test_run_build(self):
        exec.stages(
            TestHelpReturnCode.STAGES,
            keep_data=False)

class TestHelpStandardOut(unittest.TestCase):
    """
    Tests that the calls to CLI's help produce the output that users expect.
    """

    def test_first(self):
        actual = subprocess.check_output(["starfish", "--help"])
        expected = """Usage: starfish [OPTIONS] COMMAND [ARGS]..."""
        assert_contains(actual, expected)

    def test_second(self):
        actual = subprocess.check_output(["starfish", "detect_spots", "--help"])
        expected = """Usage: starfish detect_spots [OPTIONS] COMMAND [ARGS]..."""
        actual = actual.decode("utf-8")
        assert_contains(actual, expected)

    def test_third(self):
        actual = subprocess.check_output(["starfish", "detect_spots", "BlobDetector", "--help"])
        expected = """Usage: starfish detect_spots BlobDetector [OPTIONS]"""
        assert_contains(actual, expected)



from abc import abstractmethod
from typing import Type

import click

from starfish.codebook.codebook import Codebook
from starfish.intensity_table.intensity_table import IntensityTable
from starfish.pipeline.algorithmbase import AlgorithmBase
from starfish.pipeline.pipelinecomponent import PipelineComponent


COMPONENT_NAME = "decode"


class Decoder(PipelineComponent):
    @classmethod
    def pipeline_component_type_name(cls) -> str:
        return COMPONENT_NAME

    @classmethod
    def _cli_run(cls, ctx, instance):
        table = ctx.obj["intensities"]
        codes = ctx.obj["codebook"]
        output = ctx.obj["output"]
        intensities = instance.run(table, codes)
        intensities.save(output)

    @staticmethod
    @click.group(COMPONENT_NAME)
    @click.option("-i", "--input", required=True, type=click.Path(exists=True))
    @click.option("-o", "--output", required=True)
    @click.option("--codebook", required=True, type=click.Path(exists=True))
    @click.pass_context
    def _cli(ctx, input, output, codebook):
        """assign genes to spots"""
        ctx.obj = dict(
            component=Decoder,
            input=input,
            output=output,
            intensities=IntensityTable.load(input),
            codebook=Codebook.from_json(codebook),
        )


class DecoderAlgorithmBase(AlgorithmBase):
    @classmethod
    def get_pipeline_component_class(cls) -> Type[PipelineComponent]:
        return Decoder

    @abstractmethod
    def run(self, intensities: IntensityTable, codebook: Codebook, *args):
        """Performs decoding on the spots found, using the codebook specified."""
        raise NotImplementedError()



import numpy as np

from starfish.image._filter.match_histograms import MatchHistograms
from starfish.imagestack.imagestack import ImageStack
from starfish.types import Axes


def test_match_histograms():
    linear_gradient = np.linspace(0, 0.5, 2000, dtype=np.float32)
    image = linear_gradient.reshape(2, 4, 5, 5, 10)

    # linear_gradient = np.linspace(0, 1, 10)[::-1]
    # grad = np.repeat(linear_gradient[np.newaxis, :], 10, axis=0)
    # image2 = np.tile(grad, (1, 2, 2, 10, 10))

    # because of how the image was structured, every volume should be the same after
    # quantile normalization
    stack = ImageStack.from_numpy_array(image)
    mh = MatchHistograms({Axes.CH, Axes.ROUND})
    results = mh.run(stack)
    assert len(np.unique(results.xarray.sum(("x", "y", "z")))) == 1

    # because here we are allowing variation to persist across rounds, each
    # round within each channel should be different
    mh = MatchHistograms({Axes.CH})
    results2 = mh.run(stack)
    assert len(np.unique(results2.xarray.sum(("x", "y", "z")))) == 2

    # same as above, but verifying this functions for a different data shape (2 rounds, 4 channels)
    mh = MatchHistograms({Axes.ROUND})
    results2 = mh.run(stack)
    assert len(np.unique(results2.xarray.sum(("x", "y", "z")))) == 4



"""
This module parses and retains the extras metadata attached to TileSet extras.
"""
import warnings
from typing import Collection, Mapping, MutableMapping, Optional, Tuple

import numpy as np
from slicedimage import Tile, TileSet

from starfish.errors import DataFormatWarning
from starfish.imagestack.dataorder import AXES_DATA
from starfish.imagestack.parser import TileCollectionData, TileData, TileKey
from starfish.types import Axes, Coordinates, Number


class _TileSetConsistencyDetector:
    """
    This class tracks all the tile shapes and the dtypes seen thus far during the decode of a
    :py:class:`slicedimage.TileSet`.  If the shapes are not all identical, it will trigger a
    ValueError.  If the kind of dtypes are not all identical, it will trigger a TypeError.
    Additionally, if the dtypes are all of the same kind, but not all of the same size, it will
    trigger a :py:class:`starfish.errors.DataFormatWarning`.
    """
    def __init__(self) -> None:
        self.tile_shape: Optional[Mapping[Axes, int]] = None
        self.kind = None
        self.dtype_size = None

    def report_tile_shape(
            self,
            r: int, ch: int, zplane: int,
            tile_shape: Mapping[Axes, int]) -> None:
        """As each tile is parsed, the tile shape should be reported via this method.  If an
        inconsistency is detected, a ValueError exception will be raised.

        Parameters
        ----------
        r, ch, zplane : int
            The indices of the tile, which is used to generate the exception text.
        tile_shape : Mapping[Axes, int]
            The shape of the tile, mapping from Axes to its size.
        """
        if self.tile_shape is not None and self.tile_shape != tile_shape:
            raise ValueError(
                f"Tile (R: {r} C: {ch} Z: {zplane}) has shape {tile_shape}, which is different "
                f"from one or more of the other tiles.  Starfish does not support tiles that are "
                f"not identical in shape."
            )
        self.tile_shape = tile_shape

    def report_dtype(self, r: int, ch: int, zplane: int, dtype) -> None:
        """As each tile is parsed, the tile data type should be reported via this method.  If an
        inconsistency in the data type kind is is detected, a TypeError exception will be raised.
        If the data type are of the same kind but not the same size, it will trigger a
        :py:class:`starfish.errors.DataFormatWarning`

        Parameters
        ----------
        r, ch, zplane : int
            The indices of the tile, which is used to generate the exception text.
        dtype :
            The data type of the tile.
        """
        if self.kind is not None and self.kind != dtype.kind:
            raise TypeError("All tiles should have the same kind of dtype")
        if self.dtype_size is not None and self.dtype_size != dtype.itemsize:
            warnings.warn(
                f"Tile (R: {r} C: {ch} Z: {zplane}) has dtype {dtype}, which is different from one "
                f"or more of the other tiles.",
                DataFormatWarning)

        self.kind = dtype.kind
        self.dtype_size = dtype.itemsize


class SlicedImageTile(TileData):
    """
    This wraps a :py:class:`slicedimage.Tile`.  The difference between this and
    :py:class:`slicedimage.Tile` is that this class does cache the image data upon load.  It is
    therefore incumbent on the consumers of these objects to discard them as soon as it is
    reasonable to do so to free up memory.
    """
    def __init__(
            self,
            wrapped_tile: Tile,
            expectations: _TileSetConsistencyDetector,
            r: int,
            ch: int,
            zplane: int,
    ) -> None:
        self._wrapped_tile = wrapped_tile
        self._expectations = expectations
        self._r = r
        self._ch = ch
        self._zplane = zplane
        self._numpy_array: np.ndarray = None

    def _load(self):
        if self._numpy_array is not None:
            return
        self._numpy_array = self._wrapped_tile.numpy_array
        self._expectations.report_dtype(self._r, self._ch, self._zplane, self._numpy_array.dtype)

    @property
    def tile_shape(self) -> Mapping[Axes, int]:
        self._load()
        raw_tile_shape = self._numpy_array.shape
        assert len(raw_tile_shape) == 2
        tile_shape = {Axes.Y: raw_tile_shape[0], Axes.X: raw_tile_shape[1]}
        self._expectations.report_tile_shape(self._r, self._ch, self._zplane, tile_shape)
        return tile_shape

    @property
    def numpy_array(self) -> np.ndarray:
        self._load()
        return self._numpy_array

    @property
    def coordinates(self) -> Mapping[Coordinates, Tuple[Number, Number]]:
        return {
            Coordinates(coordinate_name): coordinate_value
            for coordinate_name, coordinate_value in self._wrapped_tile.coordinates.items()
        }

    @property
    def selector(self) -> Mapping[Axes, int]:
        return {
            Axes(axis_str): index
            for axis_str, index in self._wrapped_tile.indices.items()
        }


class TileSetData(TileCollectionData):
    """
    This class presents a simpler API for accessing a TileSet and its constituent tiles.
    """
    def __init__(self, tileset: TileSet) -> None:
        self._tile_shape = tileset.default_tile_shape

        self.tiles: MutableMapping[TileKey, Tile] = dict()
        for tile in tileset.tiles():
            key = TileKey(
                round=tile.indices[Axes.ROUND],
                ch=tile.indices[Axes.CH],
                zplane=tile.indices.get(Axes.ZPLANE, 0))
            self.tiles[key] = tile

            # if we don't have the tile shape, then we peek at the tile and get its shape.
            if self._tile_shape is None:
                self._tile_shape = tile.tile_shape

        self._extras = tileset.extras
        self._expectations = _TileSetConsistencyDetector()

    def __getitem__(self, tilekey: TileKey) -> dict:
        """Returns the extras metadata for a given tile, addressed by its TileKey"""
        return self.tiles[tilekey].extras

    def keys(self) -> Collection[TileKey]:
        """Returns a Collection of the TileKey's for all the tiles."""
        return self.tiles.keys()

    @property
    def tile_shape(self) -> Mapping[Axes, int]:
        return self._tile_shape

    @property
    def extras(self) -> dict:
        """Returns the extras metadata for the TileSet."""
        return self._extras

    def get_tile_by_key(self, tilekey: TileKey) -> TileData:
        return SlicedImageTile(
            self.tiles[tilekey],
            self._expectations,
            tilekey.round, tilekey.ch, tilekey.z,
        )

    def get_tile(self, r: int, ch: int, z: int) -> TileData:
        return SlicedImageTile(
            self.tiles[TileKey(round=r, ch=ch, zplane=z)],
            self._expectations,
            r, ch, z,
        )


def parse_tileset(
        tileset: TileSet
) -> Tuple[Mapping[Axes, int], TileCollectionData]:
    """
    Parse a :py:class:`slicedimage.TileSet` for formatting into an
    :py:class:`starfish.imagestack.ImageStack`.

    Parameters:
    -----------
    tileset : TileSet
        The tileset to parse.

    Returns:
    --------
    Tuple[Tuple[int, int], TileSetData] :
        A tuple consisting of the following:
            1. The (y, x) size of each tile.
            2. A :py:class:`starfish.imagestack.tileset.TileSetData` that can be queried to obtain
               the image data and extras metadata of each tile, as well as the extras metadata of
               the entire :py:class:`slicedimage.TileSet`.
    """
    tile_data = TileSetData(tileset)

    tile_shape = tileset.default_tile_shape

    # if we don't have the tile shape, then we peek at the first tile and get its shape.
    if tile_shape is None:
        tile_key = next(iter(tile_data.keys()))
        tile = tile_data.get_tile_by_key(tile_key)
        tile_shape = tile.tile_shape

    return (
        tile_shape,
        tile_data,
    )


def _get_dimension_size(tileset: TileSet, dimension: Axes):
    axis_data = AXES_DATA[dimension]
    if dimension in tileset.dimensions or axis_data.required:
        return tileset.get_dimension_shape(dimension)
    return 1



import numpy as np
from skimage.feature import register_translation
from skimage.transform._geometric import SimilarityTransform

from starfish.image._learn_transform.transforms_list import TransformsList
from starfish.imagestack.imagestack import ImageStack
from starfish.types import Axes, TransformType
from starfish.util import click
from ._base import LearnTransformBase


class Translation(LearnTransformBase):

    def __init__(self, reference_stack: ImageStack, axes: Axes, upsampling: int=1):
        """
        Parameters
        ----------
        axes:
            The axes {r, ch, zplane} to iterate over
        reference_stack: ImageStack
            The target image used in skimage.feature.register_translation
        upsampling: int
            upsampling factor (default=1). See
            http://scikit-image.org/docs/dev/api/skimage.feature.html#skimage.feature.register_translation
            for an explanation of this parameter.
        """
        self.upsampling = upsampling
        self.axes = axes
        self.reference_stack = reference_stack

    def run(self, stack: ImageStack, verbose: bool=False, *args) -> TransformsList:
        """
        Iterate over the given axes of an ImageStack and learn the Similarity transform
        based off the reference_stack passed into :py:class:`Translation`'s constructor.
        Only supports 2d data.


        Parameters
        ----------
        stack : ImageStack
            Stack to calculate the transforms on.
        verbose : bool
            if True, report on transformation progress (default = False)

        Returns
        -------
        List[Tuple[Mapping[Axes, int], SimilarityTransform]] :
            A list of tuples containing axes of the Imagestack and associated
            transform to apply.
        """

        transforms = TransformsList()
        reference_image = np.squeeze(self.reference_stack.xarray)
        for a in stack.axis_labels(self.axes):
            target_image = np.squeeze(stack.sel({self.axes: a}).xarray)
            if len(target_image.shape) != 2:
                raise ValueError(
                    f"Only axes: {self.axes.value} can have a length > 1, "
                    f"please use the MaxProj filter."
                )

            shift, error, phasediff = register_translation(src_image=target_image,
                                                           target_image=reference_image,
                                                           upsample_factor=self.upsampling)
            if verbose:
                print(f"For {self.axes}: {a}, Shift: {shift}, Error: {error}")
            selectors = {self.axes: a}
            # reverse shift because SimilarityTransform stores in y,x format
            shift = shift[::-1]
            transforms.append(selectors,
                              TransformType.SIMILARITY,
                              SimilarityTransform(translation=shift))

        return transforms

    @staticmethod
    @click.command("Translation")
    @click.option("--reference-stack", required=True, type=click.Path(exists=True),
                  help="The image to align the input ImageStack to.")
    @click.option("--axes", default="r", type=str, help="The axes to iterate over.")
    @click.option("--upsampling", default=1, type=int, help="Upsampling factor.")
    @click.pass_context
    def _cli(ctx, reference_stack, axes, upsampling):
        ctx.obj["component"]._cli_run(ctx, Translation(
            reference_stack=ImageStack.from_path_or_url(reference_stack),
            axes=Axes(axes), upsampling=upsampling))



from copy import deepcopy
from typing import Union

import numpy as np
import xarray as xr
from skimage import transform
from skimage.transform._geometric import GeometricTransform
from tqdm import tqdm

from starfish.config import StarfishConfig
from starfish.image._apply_transform._base import ApplyTransformBase
from starfish.image._learn_transform.transforms_list import TransformsList
from starfish.imagestack.imagestack import ImageStack
from starfish.types import Axes
from starfish.util import click


class Warp(ApplyTransformBase):
    """Class that applies a list of arbitrary skimage GeometricTransforms to an ImageStack
     using skimage.transform.warp"""

    def run(self, stack: ImageStack, transforms_list: TransformsList,
            in_place: bool=False, verbose: bool=False, *args, **kwargs) -> ImageStack:
        """Applies a list of transformations to an ImageStack

        Parameters
        ----------
        stack : ImageStack
            Stack to be transformed.
        transforms_list: TransformsList
            The list of transform objects to apply to the ImageStack.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on transformation progress (default = False)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of the transforms as a new stack.
            Otherwise return the original stack.
        """
        if not in_place:
            # create a copy of the ImageStack, call apply on that stack with in_place=True
            image_stack = deepcopy(stack)
            return self.run(image_stack, transforms_list, in_place=True, **kwargs)
        if verbose and StarfishConfig().verbose:
            transforms_list.transforms = tqdm(transforms_list.transforms)
        all_axes = {Axes.ROUND, Axes.CH, Axes.ZPLANE}
        for selector, _, transformation_object in transforms_list.transforms:
            other_axes = all_axes - set(selector.keys())
            # iterate through remaining axes
            for axes in stack._iter_axes(other_axes):
                # combine all axes data to select one tile
                selector.update(axes)  # type: ignore
                selected_image, _ = stack.get_slice(selector)
                warped_image = warp(selected_image, transformation_object, **kwargs
                                    ).astype(np.float32)
                stack.set_slice(selector, warped_image)
        return stack

    @staticmethod
    @click.command("Warp")
    @click.pass_context
    def _cli(ctx):
        ctx.obj["component"]._cli_run(ctx, Warp())


def warp(image: Union[xr.DataArray, np.ndarray],
         transformation_object: GeometricTransform,
         **kwargs
         ) -> np.ndarray:
    """ Wrapper around skimage.transform.warp. Warps an image according to a
    given coordinate transformation.

    Parameters
    ----------
    image: np.ndarray
        The image to be transformed
    transformation_object: skimage.transform.GeometricTransform
        The transformation object to apply.

    Returns
    -------
    np.ndarray:
        the warped image.
    """
    return transform.warp(image, transformation_object, **kwargs)



import numpy as np

from starfish import ImageStack
from starfish.image._filter.linear_unmixing import LinearUnmixing

def setup_linear_unmixing_test():
    """
        Create the image stack, coeff matrix, and reference result
        for the linear unmixing test

    """
    # Create image
    r, c, z, y, x = 2, 3, 6, 5, 4
    im = np.ones((r, c, z, y, x), dtype=np.float32)
    stack = ImageStack.from_numpy_array(im)

    # Create coefficients matrix
    coeff_mat = np.array([[1, 0, 0], [-0.25, 1, -0.25], [0, 0, 1]])

    # Create reference result
    ref_result = np.ones((r, c, z, y, x))
    ref_result[:, 1, ...] = 0.5 * np.ones((z, y, x))

    return stack, coeff_mat, ref_result

def test_linear_unmixing():
    """ Test the linear unmixing filter """

    stack, coeff_mat, ref_result = setup_linear_unmixing_test()

    filter_unmix = LinearUnmixing(coeff_mat=coeff_mat)
    stack2 = filter_unmix.run(stack, in_place=False, verbose=False, n_processes=1)

    assert np.all(ref_result == stack2.xarray.values)



import random
from collections import OrderedDict
from typing import Union

import numpy as np

from starfish import IntensityTable
from starfish.intensity_table import intensity_table_coordinates
from starfish.test import test_utils
from starfish.types import Axes, Coordinates, Features, PhysicalCoordinateTypes

NUMBER_SPOTS = 10


def physical_cord_to_pixel_value(physical_coord: float,
                                 physical_pixel_size: Union[float, int],
                                 coordinates_at_pixel_offset_0: int):

    return (physical_coord - coordinates_at_pixel_offset_0) / physical_pixel_size


def test_tranfering_physical_coords_to_intensity_table():
    stack_shape = OrderedDict([(Axes.ROUND, 3), (Axes.CH, 2),
                               (Axes.ZPLANE, 1), (Axes.Y, 50), (Axes.X, 40)])

    physical_coords = OrderedDict([(PhysicalCoordinateTypes.X_MIN, 1),
                                   (PhysicalCoordinateTypes.X_MAX, 2),
                                   (PhysicalCoordinateTypes.Y_MIN, 4),
                                   (PhysicalCoordinateTypes.Y_MAX, 6),
                                   (PhysicalCoordinateTypes.Z_MIN, 1),
                                   (PhysicalCoordinateTypes.Z_MAX, 3)])

    stack = test_utils.imagestack_with_coords_factory(stack_shape, physical_coords)
    codebook = test_utils.codebook_array_factory()

    intensities = IntensityTable.synthetic_intensities(
        codebook,
        num_z=stack_shape[Axes.ZPLANE],
        height=stack_shape[Axes.Y],
        width=stack_shape[Axes.X],
        n_spots=NUMBER_SPOTS
    )

    intensities = intensity_table_coordinates.\
        transfer_physical_coords_from_imagestack_to_intensity_table(stack, intensities)

    # Assert that new cords were added
    xc = intensities.coords[Coordinates.X]
    yc = intensities.coords[Coordinates.Y]
    zc = intensities.coords[Coordinates.Z]
    assert xc.size == NUMBER_SPOTS
    assert yc.size == NUMBER_SPOTS
    assert zc.size == NUMBER_SPOTS

    # Assert that the physical coords align with their corresponding pixel coords
    for spot in xc.features:
        pixel_x = spot[Axes.X.value].data
        physical_x = stack.xarray[Coordinates.X.value][pixel_x]
        assert np.isclose(spot[Coordinates.X.value], physical_x)

    for spot in yc.features:
        pixel_y = spot[Axes.Y.value].data
        physical_y = stack.xarray[Coordinates.Y.value][pixel_y]
        assert np.isclose(spot[Coordinates.Y.value], physical_y)

    # Assert that zc value is middle of z range
    for spot in zc.features:
        z_plane = spot[Axes.ZPLANE.value].data
        physical_z = stack.xarray[Coordinates.Z.value][z_plane]
        assert np.isclose(spot[Coordinates.Z.value], physical_z)


def test_tranfering_physical_coords_to_expression_matrix():
    stack_shape = OrderedDict([(Axes.ROUND, 3), (Axes.CH, 2),
                               (Axes.ZPLANE, 1), (Axes.Y, 50), (Axes.X, 40)])

    physical_coords = OrderedDict([(PhysicalCoordinateTypes.X_MIN, 1),
                                   (PhysicalCoordinateTypes.X_MAX, 2),
                                   (PhysicalCoordinateTypes.Y_MIN, 4),
                                   (PhysicalCoordinateTypes.Y_MAX, 6),
                                   (PhysicalCoordinateTypes.Z_MIN, 1),
                                   (PhysicalCoordinateTypes.Z_MAX, 3)])

    stack = test_utils.imagestack_with_coords_factory(stack_shape, physical_coords)
    codebook = test_utils.codebook_array_factory()

    intensities = IntensityTable.synthetic_intensities(
        codebook,
        num_z=stack_shape[Axes.ZPLANE],
        height=stack_shape[Axes.Y],
        width=stack_shape[Axes.X],
        n_spots=NUMBER_SPOTS
    )

    intensities = intensity_table_coordinates. \
        transfer_physical_coords_from_imagestack_to_intensity_table(stack, intensities)

    # Check that error is thrown before target assignment
    try:
        intensities.to_expression_matrix()
    except KeyError as e:
        # Assert value error is thrown with right message
        assert e.args[0] == "IntensityTable must have 'cell_id' assignments for each cell before " \
                            "this function can be called. See starfish.TargetAssignment.Label."

    # mock out come cell_ids
    cell_ids = random.sample(range(1, 20), NUMBER_SPOTS)
    intensities[Features.CELL_ID] = (Features.AXIS, cell_ids)

    expression_matrix = intensities.to_expression_matrix()
    # Assert that coords were transferred
    xc = expression_matrix.coords[Coordinates.X]
    yc = expression_matrix.coords[Coordinates.Y]
    zc = expression_matrix.coords[Coordinates.Z]
    assert xc.size == len(set(cell_ids))
    assert yc.size == len(set(cell_ids))
    assert zc.size == len(set(cell_ids))



import copy

import numpy as np

from starfish.imagestack.imagestack import ImageStack
from starfish.types import Axes, Clip
from starfish.util.synthesize import SyntheticData


def divide(array, value):
    return array / value


def test_apply():
    """test that apply correctly applies a simple function across 2d tiles of a Stack"""
    stack = ImageStack.synthetic_stack()
    assert (stack.xarray == 1).all()
    output = stack.apply(divide, value=2)
    assert (output.xarray == 0.5).all()


def test_apply_3d():
    """test that apply correctly applies a simple function across 3d volumes of a Stack"""
    stack = ImageStack.synthetic_stack()
    assert np.all(stack.xarray == 1)
    stack.apply(divide, in_place=True, value=4,
                group_by={Axes.ROUND, Axes.CH})
    assert (stack.xarray == 0.25).all()


def test_apply_labeled_dataset():
    """
    test that apply correctly applies a simple function across starfish-generated synthetic data
    """
    original = SyntheticData().spots()
    image = original.apply(divide, value=2)
    assert np.all(image.xarray == original.xarray / 2)


def test_apply_in_place():
    """
    test that apply correctly applies a simple function across a starfish stack without modifying
    original data
    """
    image = SyntheticData().spots()
    original = copy.deepcopy(image)
    image.apply(divide, value=2, in_place=True)
    assert np.all(image.xarray == original.xarray / 2)


def test_apply_single_process():
    """test that apply correctly applies a simple function across 2d tiles of a Stack"""
    stack = ImageStack.synthetic_stack()
    assert (stack.xarray == 1).all()
    output = stack.apply(divide, value=2, n_processes=1)
    assert (output.xarray == 0.5).all()

def test_apply_clipping_methods():
    """test that apply properly clips the imagestack"""

    # create a half-valued float array
    data = np.full((2, 2, 2, 5, 5), fill_value=0.5, dtype=np.float32)
    # set one value to max
    data[1, 1, 1, 1, 1] = 1

    imagestack = ImageStack.from_numpy_array(data)

    # max value after multiplication == 2, all other values == 1
    def apply_function(x):
        return x * 2

    # clip_method 0
    # all data are clipped to 1, setting all values to 1 (np.unique(pre_scaled) == [1, 2])
    res = imagestack.apply(apply_function, clip_method=Clip.CLIP, in_place=False, n_processes=1)
    assert np.allclose(res.xarray.values, 1)

    # clip_method 1
    # all data are scaled, resulting in values being multiplied by 0.5, replicating the original
    # data
    res = imagestack.apply(
        apply_function, clip_method=Clip.SCALE_BY_IMAGE, in_place=False, n_processes=1
    )
    assert np.allclose(imagestack.xarray, res.xarray)

    # clip_method 2
    res = imagestack.apply(
        apply_function, clip_method=Clip.SCALE_BY_CHUNK, in_place=False, n_processes=1,
        group_by={Axes.CH, Axes.ROUND},
    )
    # any (round, ch) combination that was all 0.5 should now be all 1.
    assert np.allclose(res.sel({Axes.ROUND: 0, Axes.CH: 0}).xarray, 1)
    assert np.allclose(res.sel({Axes.ROUND: 1, Axes.CH: 0}).xarray, 1)
    assert np.allclose(res.sel({Axes.ROUND: 0, Axes.CH: 1}).xarray, 1)

    # the specific (round, ch) combination with the single "1" value should be scaled, and due to
    # construction, look like the original data.
    assert np.allclose(
        res.sel({Axes.ROUND: 1, Axes.CH: 1}).xarray,
        imagestack.sel({Axes.ROUND: 1, Axes.CH: 1}).xarray
    )



from abc import abstractmethod
from typing import Type

from starfish.image._learn_transform.transforms_list import TransformsList
from starfish.imagestack.imagestack import ImageStack
from starfish.pipeline import PipelineComponent
from starfish.pipeline.algorithmbase import AlgorithmBase
from starfish.util import click


COMPONENT_NAME = "apply_transform"


class ApplyTransform(PipelineComponent):

    @classmethod
    def pipeline_component_type_name(cls) -> str:
        return COMPONENT_NAME

    @classmethod
    def _cli_run(cls, ctx, instance):
        output = ctx.obj["output"]
        stack = ctx.obj["stack"]
        transformation_list = ctx.obj["transformation_list"]
        transformed = instance.run(stack, transformation_list)
        transformed.export(output)

    @staticmethod
    @click.group(COMPONENT_NAME)
    @click.option("-i", "--input", type=click.Path(exists=True))
    @click.option("-o", "--output", required=True)
    @click.option("--transformation-list", required=True, type=click.Path(exists=True),
                  help="The list of transformations to apply to the ImageStack.")
    @click.pass_context
    def _cli(ctx, input, output, transformation_list):
        print("Applying Transform to images...")
        ctx.obj = dict(
            component=ApplyTransform,
            input=input,
            output=output,
            stack=ImageStack.from_path_or_url(input),
            transformation_list=TransformsList.from_json(transformation_list)
        )


class ApplyTransformBase(AlgorithmBase):
    @classmethod
    def get_pipeline_component_class(cls) -> Type[PipelineComponent]:
        return ApplyTransform

    @abstractmethod
    def run(self, stack, transforms_list, *args) -> ImageStack:
        """Performs registration on the stack provided."""
        raise NotImplementedError()



# deal with numpy import warnings due to cython
# See: https://stackoverflow.com/questions/40845304/
#      runtimewarning-numpy-dtype-size-changed-may-indicate-binary-incompatibility)
import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")  # noqa
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")  # noqa

import pkg_resources

# image processing methods and objects
from . import image
# spot detection and manipulation
from . import spots
# display images and spots
from ._display import display
# generated version number and commit hash
from ._version import get_versions
# top-level objects
from .codebook.codebook import Codebook
from .experiment.experiment import Experiment, FieldOfView
from .imagestack.imagestack import ImageStack
from .intensity_table.intensity_table import IntensityTable
from .starfish import starfish


# NOTE: if we move to python 3.7, we can produce this value at call time via __getattr__
__version__ = get_versions()['version']
__is_release_tag__ = None
if "+" not in str(__version__):
    __is_release_tag__ = f"Release: {__version__}"
del get_versions

if __name__ == "__main__":
    starfish()



import json
from typing import List, Mapping, Tuple

import numpy as np
from skimage.transform._geometric import GeometricTransform, SimilarityTransform
from slicedimage.io import resolve_path_or_url

from starfish.config import StarfishConfig
from starfish.types import Axes, TransformType


transformsTypeMapping = {
    TransformType.SIMILARITY: SimilarityTransform
}


class TransformsList:
    """Simple list wrapper class for storing a list of transformation
    objects to apply to an Imagestack"""

    def __init__(self,
                 transforms_list: List[Tuple[Mapping[Axes, int],
                                             TransformType,
                                             GeometricTransform]] = None
                 ):
        """
        Parameters
        ----------
        transforms_list: List[Tuple[Mapping[Axes, int], TransformType, GeometricTransform]]
            A list of tuples containing axes of an Imagestack and associated
            transform to apply.
        """

        if transforms_list:
            self.transforms = transforms_list
        else:
            self.transforms: List[Tuple[Mapping[Axes, int],
                                        TransformType, GeometricTransform]] = list()

    def append(self,
               selectors: Mapping[Axes, int],
               transform_type: TransformType,
               transform_object: GeometricTransform
               ) -> None:
        """
        Adds a new GoemetricTransform object to the list

        Parameters
        ----------
        transform_object:
            The TransformationObject to add
        transform_type:
            The type of transform
        selectors:
            The axes associated with the transform.

        """
        self.transforms.append((selectors, transform_type, transform_object))

    def to_json(self, filename: str) -> None:
        """
        Saves the TransformsList to a json file.

        Parameters
        ----------
        filename : str
        """
        transforms_array = []
        # Need to convert Axes to str values, and TransformationObjects to array
        for selectors, transform_type, transforms_object in self.transforms:
            transforms_matrix = transforms_object.params.tolist()
            selectors_str = {k.value: v for k, v in selectors.items()}
            transforms_array.append((selectors_str, transform_type.value, transforms_matrix))
        with open(filename, 'w') as f:
            json.dump(transforms_array, f)

    @classmethod
    def from_json(cls, url_or_path: str) -> "TransformsList":
        """
        Load a TransformsList from a json file or a url pointing to such a file
        Loads slicedimage version configuration from :py:class:`starfish.config.StarfishConfig`

        Parameters
        ----------
        url_or_path : str
            Either an absolute URL or a filesystem path to a transformsList.

        Returns
        -------
        TransformsList
        """
        config = StarfishConfig()
        transforms_list: List[Tuple[Mapping[Axes, int], TransformType, GeometricTransform]] = list()
        backend, name, _ = resolve_path_or_url(url_or_path, backend_config=config.slicedimage)
        with backend.read_contextmanager(name) as fh:
            transforms_array = json.load(fh)
        for selectors_str, transform_type_str, transforms_matrix in transforms_array:
            selectors = {Axes(k): v for k, v in selectors_str.items()}
            transform_type = TransformType(transform_type_str)
            transform_object = transformsTypeMapping[transform_type](np.array(transforms_matrix))
            transforms_list.append((selectors, transform_type, transform_object))
        return cls(transforms_list)



import argparse
import json
import pathlib
import posixpath

import requests

from starfish import Experiment, FieldOfView
from starfish.util.argparse import FsExistsType


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--primary-name", default="primary_images.json")
    parser.add_argument("experiment_url")
    parser.add_argument("output_dir", type=FsExistsType())
    args = parser.parse_args()

    # save image stacks locally
    exp = Experiment.from_json(posixpath.join(args.experiment_url, "experiment.json"))
    for fov in exp.fovs():
        fov_dir = pathlib.Path(args.output_dir, fov.name)
        fov_dir.mkdir()
        fov.get_image(FieldOfView.PRIMARY_IMAGES).export(str(fov_dir / args.primary_name))
        for image_type in fov.image_types:
            if image_type == FieldOfView.PRIMARY_IMAGES:
                continue
            fov.get_image(image_type).export(str(fov_dir / f"{image_type}.json"))

    # get codebook from url and save locally to tmp dir
    codebook = requests.get(posixpath.join(args.experiment_url, "codebook.json"))
    data = codebook.json()
    with open(pathlib.Path(args.output_dir, 'codebook.json'), 'w') as f:
        json.dump(data, f)

    # get json files from url and save locally to tmp dir
    for filename in ("experiment.json", args.primary_name, "nuclei.json", "dots.json"):
        obj = requests.get(posixpath.join(args.experiment_url, filename))
        # download file if it exists
        if obj.status_code == 200:
            data = obj.json()
            with open(pathlib.Path(args.output_dir, filename), 'w') as f:
                json.dump(data, f)



from typing import Optional, Tuple, Union


from starfish.types import (
    Number,
)


def _calculate_physical_pixel_size(coord_min: Number, coord_max: Number, num_pixels: int) -> Number:
    """Calculate the size of a pixel in physical space"""
    return (coord_max - coord_min) / num_pixels


def _pixel_offset_to_physical_coordinate(
        physical_pixel_size: Number,
        pixel_offset: Optional[int],
        coordinates_at_pixel_offset_0: Number,
        dimension_size: int,
) -> Number:
    """Calculate the physical pixel value at the given index"""
    if pixel_offset:
        # Check for negative index
        if pixel_offset < 0:
            pixel_offset = pixel_offset + dimension_size
        return (physical_pixel_size * pixel_offset) + coordinates_at_pixel_offset_0
    return coordinates_at_pixel_offset_0


def recalculate_physical_coordinate_range(
        coord_min: float,
        coord_max: float,
        dimension_size: int,
        indexer: Union[int, slice],
) -> Tuple[Number, Number]:
    """Given the dimension size and pixel coordinate indexes calculate the corresponding
    coordinates in physical space

    Parameters
    ----------
    coord_min: float
        the minimum physical coordinate value

    coord_max: float
        the maximum physical coordinate value

    dimension_size: int
        The size (number of pixels) to use to calculate physical_pixel_size

    indexer: (int/slice)
        The pixel index or range to calculate.

    Returns
    -------
    The new min and max physical coordinate values of the given dimension
    """
    physical_pixel_size = _calculate_physical_pixel_size(coord_min, coord_max, dimension_size)
    min_pixel_index = indexer if isinstance(indexer, int) else indexer.start
    max_pixel_index = indexer if isinstance(indexer, int) else indexer.stop
    # Add one to max pixel index to get end of pixel
    max_pixel_index = max_pixel_index + 1 if max_pixel_index else dimension_size
    new_min = _pixel_offset_to_physical_coordinate(
        physical_pixel_size, min_pixel_index, coord_min, dimension_size)
    new_max = _pixel_offset_to_physical_coordinate(
        physical_pixel_size, max_pixel_index, coord_min, dimension_size)
    return new_min, new_max


def get_physical_coordinates_of_z_plane(zrange: Tuple[float, float]):
    """Calculate the midpoint of the given zrange."""
    physical_z = (zrange[1] - zrange[0]) / 2 + zrange[0]
    return physical_z



from itertools import product
from typing import Iterator, Mapping, Sequence, Tuple

from starfish.types import Axes


def join_axes_labels(
        axes_order: Sequence[Axes],
        *,
        rounds: Sequence[int],
        chs: Sequence[int],
        zplanes: Sequence[int],
) -> Sequence[Tuple[Axes, Sequence[int]]]:
    """
    Given a sequence of axes and their labels, return a sequence of tuples of axes and its
    respective labels.

    For example, if axes_sequence is (ROUND, CH, Z) and each axes has labels [0, 1], return
    ((ROUND, [0, 1]), (CH, [0, 1]), (Z, [0, 1]).
    """
    axes_mapping = {
        Axes.ROUND: rounds,
        Axes.CH: chs,
        Axes.ZPLANE: zplanes,
    }

    return [
        (axes, axes_mapping[axes])
        for axes in axes_order
    ]


def ordered_iterator(
        axes_labels: Sequence[Tuple[Axes, Sequence[int]]]
) -> Iterator[Mapping[Axes, int]]:
    """
    Given a sequence of tuples of axes and its respective sequence of labels, return an iterator
    that steps through all the possible points in the space.  The sequence is ordered from the
    slowest varying axes to the fastest varying axes.
    """
    for tpl in product(*[labels for _, labels in axes_labels]):
        yield dict(zip((index for index, _ in axes_labels), tpl))



from collections import OrderedDict
from typing import Sequence, Tuple

import numpy as np
import xarray as xr
from scipy.ndimage.filters import gaussian_filter

from starfish import Codebook, ImageStack
from starfish.imagestack import physical_coordinate_calculator
from starfish.types import Axes, Coordinates, Features, PhysicalCoordinateTypes


def imagestack_with_coords_factory(stack_shape: OrderedDict, coords: OrderedDict) -> ImageStack:
    """
    Create an ImageStack of given shape and assigns the given x,y,z
    min/max physical coordinates to each tile.

    Parameters
    ----------
    stack_shape: OrderedDict
        Dict[Axes, int] defining the size of each dimension for an ImageStack

    coords: OrderedDict
        Dict[PhysicalCoordinateTypes, float] defining the min/max values of physical
        coordinates to assign to the Imagestack
    """

    stack = ImageStack.synthetic_stack(num_round=stack_shape[Axes.ROUND],
                                       num_ch=stack_shape[Axes.CH],
                                       num_z=stack_shape[Axes.ZPLANE],
                                       tile_height=stack_shape[Axes.Y],
                                       tile_width=stack_shape[Axes.X])

    stack.xarray[Coordinates.X.value] = xr.DataArray(
        np.linspace(coords[PhysicalCoordinateTypes.X_MIN], coords[PhysicalCoordinateTypes.X_MAX],
                    stack.xarray.sizes[Axes.X.value]), dims=Axes.X.value)

    stack.xarray[Coordinates.Y.value] = xr.DataArray(
        np.linspace(coords[PhysicalCoordinateTypes.Y_MIN], coords[PhysicalCoordinateTypes.Y_MAX],
                    stack.xarray.sizes[Axes.Y.value]), dims=Axes.Y.value)

    z_coord = physical_coordinate_calculator.\
        get_physical_coordinates_of_z_plane((coords[PhysicalCoordinateTypes.Z_MIN],
                                             coords[PhysicalCoordinateTypes.Z_MAX]))

    stack.xarray[Coordinates.Z.value] = xr.DataArray(np.zeros(
        stack.xarray.sizes[Axes.ZPLANE.value]),
        dims=Axes.ZPLANE.value)

    for z in stack.axis_labels(Axes.ZPLANE):
        stack.xarray[Coordinates.Z.value].loc[z] = z_coord

    return stack


def codebook_array_factory() -> Codebook:
    """
    Codebook with two codewords describing an experiment with three channels and two imaging rounds.
    Both codes have two "on" channels.
    """
    data = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 2, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
    ]
    return Codebook.from_code_array(data)


def _create_dataset(
    pixel_dimensions: Tuple[int, int, int],
    spot_coordinates: Sequence[Tuple[int, int, int]],
    codebook: Codebook
) -> ImageStack:
    """
    creates a numpy array containing one spot per codebook entry at spot_coordinates. length of
    spot_coordinates must therefore match the number of codes in Codebook.
    """
    assert len(spot_coordinates) == codebook.sizes[Features.TARGET]

    data_shape = (
        codebook.sizes[Axes.ROUND.value],
        codebook.sizes[Axes.CH.value],
        *pixel_dimensions
    )
    imagestack_data = np.zeros((data_shape), dtype=np.float32)

    for ((z, y, x), f) in zip(spot_coordinates, range(codebook.sizes[Features.TARGET])):
        imagestack_data[:, :, z, y, x] = codebook[f].transpose(Axes.ROUND.value, Axes.CH.value)

    # blur with a small non-isotropic kernel TODO make kernel smaller.
    imagestack_data = gaussian_filter(imagestack_data, sigma=(0, 0, 0.7, 1.5, 1.5))
    return ImageStack.from_numpy_array(imagestack_data)


def two_spot_one_hot_coded_data_factory() -> Tuple[Codebook, ImageStack, float]:
    """
    Produce a 2-channel 2-round Codebook with two codes and an ImageStack containing one spot from
    each code. The spots do not overlap and the data are noiseless.

    The encoding of this data is similar to that used in In-situ Sequencing, FISSEQ,
    BaristaSeq, STARMAP, MExFISH, or SeqFISH.

    Returns
    -------
    Codebook :
        codebook containing codes that match the data
    ImageStack :
        noiseless ImageStack containing one spot per code in codebook
    float :
        the maximum intensity found in the created ImageStack

    """

    codebook_data = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
    ]
    codebook = Codebook.from_code_array(codebook_data)

    imagestack = _create_dataset(
        pixel_dimensions=(10, 100, 100),
        spot_coordinates=((4, 10, 90), (5, 90, 10)),
        codebook=codebook
    )

    max_intensity = np.max(imagestack.xarray.values)

    return codebook, imagestack, max_intensity


def two_spot_sparse_coded_data_factory() -> Tuple[Codebook, ImageStack, float]:
    """
    Produce a 3-channel 3-round Codebook with two codes and an ImageStack containing one spot from
    each code. The spots do not overlap and the data are noiseless.

    These spots display sparsity in both rounds and channels, similar to the sparse encoding of
    MERFISH

    Returns
    -------
    ImageStack :
        noiseless ImageStack containing two spots

    """

    codebook_data = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 2, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 2, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
    ]
    codebook = Codebook.from_code_array(codebook_data)

    imagestack = _create_dataset(
        pixel_dimensions=(10, 100, 100),
        spot_coordinates=((4, 10, 90), (5, 90, 10)),
        codebook=codebook
    )

    max_intensity = np.max(imagestack.xarray.values)

    return codebook, imagestack, max_intensity


def two_spot_informative_blank_coded_data_factory() -> Tuple[Codebook, ImageStack, float]:
    """
    Produce a 4-channel 2-round Codebook with two codes and an ImageStack containing one spot from
    each code. The spots do not overlap and the data are noiseless.

    The encoding of this data is essentially a one-hot encoding, but where one of the channels is a
    intentionally and meaningfully "blank".

    Returns
    -------
    Codebook :
        codebook containing codes that match the data
    ImageStack :
        noiseless ImageStack containing one spot per code in codebook
    float :
        the maximum intensity found in the created ImageStack

    """

    codebook_data = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                # round 3 is blank and channel 3 is not used
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                # round 0 is blank and channel 0 is not used
                {Axes.ROUND.value: 1, Axes.CH.value: 3, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 2, Axes.CH.value: 2, Features.CODE_VALUE: 1},
            ],
            Features.TARGET: "GENE_B"
        },
    ]
    codebook = Codebook.from_code_array(codebook_data)

    imagestack = _create_dataset(
        pixel_dimensions=(10, 100, 100),
        spot_coordinates=((4, 10, 90), (5, 90, 10)),
        codebook=codebook
    )

    max_intensity = np.max(imagestack.xarray.values)

    return codebook, imagestack, max_intensity



#!/usr/bin/env python
import cProfile
import subprocess
import sys
from pstats import Stats

import pkg_resources

from starfish.experiment.builder.cli import build as build_cli
from starfish.image import (
    ApplyTransform,
    Filter,
    LearnTransform,
    Segmentation,
)
from starfish.spacetx_format.cli import validate as validate_cli
from starfish.spots import (
    Decoder,
    PixelSpotDecoder,
    SpotFinder,
    TargetAssignment,
)
from starfish.util import click


PROFILER_KEY = "profiler"
"""This is the dictionary key we use to attach the profiler to pass to the resultcallback."""
PROFILER_LINES = 15
"""This is the number of profiling rows to dump when --profile is enabled."""


def art_string():
    return r"""
         _              __ _     _
        | |            / _(_)   | |
     ___| |_ __ _ _ __| |_ _ ___| |__
    / __| __/ _` | '__|  _| / __| '_  `
    \__ \ || (_| | |  | | | \__ \ | | |
    |___/\__\__,_|_|  |_| |_|___/_| |_|

    """

@click.group()
@click.option("--profile", is_flag=True)
@click.pass_context
def starfish(ctx, profile):
    """
    standardized analysis pipeline for image-based transcriptomics
    see: https://spacetx-starfish.readthedocs.io for more information.
    """
    art = art_string()
    print_art = True
    sub = ctx.command.get_command(ctx, ctx.invoked_subcommand)
    if hasattr(sub, "no_art"):
        print_art = not sub.no_art
    if print_art:
        print(art)
    if profile:
        profiler = cProfile.Profile()
        profiler.enable()

        def print_profile():
            stats = Stats(profiler)
            stats.sort_stats('tottime').print_stats(PROFILER_LINES)

        ctx.call_on_close(print_profile)


@starfish.command()
def version():
    from starfish import __version__
    print(__version__)
version.no_art = True  # type: ignore


@starfish.group()
def util():
    """
    house-keeping commands for the starfish library
    """
    pass

@util.command()
def install_strict_dependencies():
    """
    warning! updates different packages in your local installation
    """
    strict_requirements_file = pkg_resources.resource_filename(
        "starfish", "REQUIREMENTS-STRICT.txt")
    subprocess.check_call([
        sys.executable, "-m", "pip", "install", "-r", strict_requirements_file
    ])


# Pipelines
starfish.add_command(LearnTransform._cli)  # type: ignore
starfish.add_command(ApplyTransform._cli)  # type: ignore
starfish.add_command(Filter._cli)  # type: ignore
starfish.add_command(PixelSpotDecoder._cli)
starfish.add_command(SpotFinder._cli)  # type: ignore
starfish.add_command(Segmentation._cli)  # type: ignore
starfish.add_command(TargetAssignment._cli)  # type: ignore
starfish.add_command(Decoder._cli)  # type: ignore

# Other
starfish.add_command(build_cli)  # type: ignore
starfish.add_command(validate_cli)  # type: ignore



from copy import deepcopy
from functools import lru_cache

import numpy as np

from starfish.display import (
    _mask_low_intensity_spots,
    _max_intensity_table_maintain_dims,
    _spots_to_markers,
)
from starfish.intensity_table.intensity_table import IntensityTable
from starfish.test import test_utils
from starfish.types import Axes, Features


@lru_cache(maxsize=1)
def testing_data():
    np.random.seed(1)
    codebook = test_utils.codebook_array_factory()
    num_z, height, width = 3, 4, 5
    intensities = IntensityTable.synthetic_intensities(
        codebook,
        num_z=num_z,
        height=height,
        width=width,
        n_spots=4
    )
    return intensities


def test_max_intensity_maintains_each_combination_of_dimensions():
    # Frozen(OrderedDict([('features', 4), ('c', 3), ('r', 2)]))
    intensities = testing_data()
    results = _max_intensity_table_maintain_dims(intensities, {Axes.CH})
    assert results.shape == (4, 1, 2)
    results = _max_intensity_table_maintain_dims(intensities, {Axes.ROUND})
    assert results.shape == (4, 3, 1)
    results = _max_intensity_table_maintain_dims(intensities, {Axes.ROUND, Axes.CH})
    assert results.shape == (4, 1, 1)
    results = _max_intensity_table_maintain_dims(intensities, {Features.AXIS, Axes.CH})
    assert results.shape == (1, 1, 2)
    results = _max_intensity_table_maintain_dims(
        intensities, {Axes.ROUND, Axes.CH, Features.AXIS}
    )
    assert results.shape == (1, 1, 1)


def test_mask_low_intensities_masks_preexisting_nans():
    intensities = testing_data()
    masked = _mask_low_intensity_spots(intensities, intensity_threshold=0)
    # intensity_threshold == 0 should mean no spots are masked
    assert np.sum(masked) == np.product(intensities.shape)

    # with c-order ravel, this nan is in the 5th position (zero-based = 4)
    intensities = deepcopy(intensities)
    intensities[0, 2, 0] = np.nan
    masked = _mask_low_intensity_spots(intensities, intensity_threshold=0)
    assert np.array_equal(np.where(~masked)[0], np.array([4]))


def test_mask_low_intensities():
    intensities = testing_data()
    masked = _mask_low_intensity_spots(intensities, intensity_threshold=0.1)
    expected = np.where(np.ravel(intensities.values >= 0.1))
    passing = np.where(masked)
    assert np.array_equal(passing, expected)


def test_spots_to_markers():
    intensities = testing_data()
    markers, sizes = _spots_to_markers(intensities)
    assert len(markers) == np.product(intensities.shape)

    # check that intensities are being flattened into markers with r rotating fastest, then c
    assert np.all(markers[:2, 2] == [0, 1])  # rounds
    assert np.all(markers[:2, 3] == [0, 0])  # channels



import typing

from starfish.types import Axes


class TileKey:
    """
    This class is used to index into the TileData class.
    """
    def __init__(self, *, round: int, ch: int, zplane: int) -> None:
        self._round = round
        self._ch = ch
        self._zplane = zplane

    @property
    def round(self) -> int:
        return self._round

    @property
    def ch(self) -> int:
        return self._ch

    @property
    def z(self) -> int:
        return self._zplane

    AXES_TO_PROPERTY_MAP = {
        Axes.ROUND: round,
        Axes.CH: ch,
        Axes.ZPLANE: z,
    }

    def __getitem__(self, item) -> int:
        """Given a index, return the corresponding value for that index for this tilekey.  For
        instance, tilekey[Axes.ROUND] returns the round for this tilekey."""
        index = typing.cast(Axes, item)
        prop: property = TileKey.AXES_TO_PROPERTY_MAP[index]  # type: ignore
        return prop.__get__(self, TileKey)

    def __eq__(self, other) -> bool:
        if not isinstance(other, TileKey):
            return False

        return self._round == other.round and self._ch == other.ch and self._zplane == other.z

    def __hash__(self) -> int:
        return int(self._round ^ self._ch ^ self._zplane)

    def __repr__(self) -> str:
        return f"(round: {self._round} ch: {self._ch} z: {self._zplane})"



"""
This module encapsulates the logic to parse a numpy array into an ImageStack.
"""
from itertools import product
from typing import (
    Any,
    Collection,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Tuple,
)

import numpy as np
import xarray as xr

from starfish.imagestack.parser import TileCollectionData, TileData, TileKey
from starfish.types import (
    Axes,
    Coordinates,
    Number,
    PHYSICAL_COORDINATE_DIMENSION,
    PhysicalCoordinateTypes,
)


class NumpyImageTile(TileData):
    """
    This is a specialization of :py:class:`starfish.imagestack.parser.TileData` for serving data
    about a single tile coming from a numpy array.
    """
    def __init__(
            self,
            data: np.ndarray,
            coordinates: Mapping[Coordinates, Tuple[Number, Number]],
            selector: Mapping[Axes, int],
    ) -> None:
        self._data = data
        self._coordinates = coordinates
        self._selector = selector

    @property
    def tile_shape(self) -> Mapping[Axes, int]:
        raw_tile_shape = self._data.shape
        assert len(raw_tile_shape) == 2
        tile_shape = {Axes.Y: raw_tile_shape[0], Axes.X: raw_tile_shape[1]}
        return tile_shape

    @property
    def numpy_array(self):
        return self._data

    @property
    def coordinates(self) -> Mapping[Coordinates, Tuple[Number, Number]]:
        return self._coordinates

    @property
    def selector(self) -> Mapping[Axes, int]:
        return self._selector


class NumpyData(TileCollectionData):
    """
    This is a specialization of :py:class:`starfish.imagestack.parser.TileCollectionData` for
    serving tile data from a numpy array.
    """
    def __init__(
            self,
            data: np.ndarray,
            index_labels: Mapping[Axes, Sequence[int]],
            coordinates: Optional[xr.DataArray],
    ) -> None:
        self.data = data
        self.index_labels = index_labels
        self.coordinates = coordinates

    def __getitem__(self, tilekey: TileKey) -> dict:
        """Returns the extras metadata for a given tile, addressed by its TileKey"""
        return {}

    def keys(self) -> Collection[TileKey]:
        """Returns a Collection of the TileKey's for all the tiles."""
        keys: MutableSequence[TileKey] = list()
        axis_names: MutableSequence[Axes] = list()
        labels: MutableSequence[Sequence[int]] = list()
        for index_name, index_labels in self.index_labels.items():
            axis_names.append(index_name)
            labels.append(index_labels)

        for indices in product(*labels):
            selector: MutableMapping[Axes, int] = dict()
            for index_name, index_value in zip(axis_names, indices):
                selector[index_name] = index_value

            keys.append(
                TileKey(
                    round=selector[Axes.ROUND],
                    ch=selector[Axes.CH],
                    zplane=selector[Axes.ZPLANE],
                )
            )

        return keys

    @property
    def tile_shape(self) -> Mapping[Axes, int]:
        return {Axes.Y: self.data.shape[-2], Axes.X: self.data.shape[-1]}

    @property
    def extras(self) -> dict:
        """Returns the extras metadata for the TileSet."""
        return {}

    def get_tile_by_key(self, tilekey: TileKey) -> TileData:
        return self.get_tile(tilekey.round, tilekey.ch, tilekey.z)

    def get_tile(self, r: int, ch: int, z: int) -> TileData:
        # find the positional r/ch/z.
        pos_r = self.index_labels[Axes.ROUND].index(r)
        pos_ch = self.index_labels[Axes.CH].index(ch)
        pos_z = self.index_labels[Axes.ZPLANE].index(z)

        selectors: Mapping[str, Any] = {
            Axes.ROUND.value: r,
            Axes.CH.value: ch,
            Axes.ZPLANE.value: z,
        }

        coordinates: MutableMapping[Coordinates, Tuple[Number, Number]] = dict()

        if self.coordinates is not None:
            for coordinate_type, min_selector_value, max_selector_value in (
                    (Coordinates.X, PhysicalCoordinateTypes.X_MIN, PhysicalCoordinateTypes.X_MAX),
                    (Coordinates.Y, PhysicalCoordinateTypes.Y_MIN, PhysicalCoordinateTypes.Y_MAX),
                    (Coordinates.Z, PhysicalCoordinateTypes.Z_MIN, PhysicalCoordinateTypes.Z_MAX),
            ):
                min_selectors = dict(selectors)
                min_selectors[PHYSICAL_COORDINATE_DIMENSION] = min_selector_value.value
                max_selectors = dict(selectors)
                max_selectors[PHYSICAL_COORDINATE_DIMENSION] = max_selector_value.value

                coordinates[coordinate_type] = (
                    self.coordinates.loc[min_selectors].item(),
                    self.coordinates.loc[max_selectors].item())
        else:
            # fake coordinates!
            coordinates[Coordinates.X] = (0.0, 0.001)
            coordinates[Coordinates.Y] = (0.0, 0.001)
            coordinates[Coordinates.Z] = (0.0, 0.001)

        return NumpyImageTile(
            self.data[pos_r, pos_ch, pos_z],
            coordinates,
            {
                Axes.ROUND: r,
                Axes.CH: ch,
                Axes.ZPLANE: z,
            },
        )



"""
This module describes the contracts to provide data to the experiment builder.
"""

from typing import Mapping, Tuple, Union

import numpy as np

from starfish.types import Axes, Coordinates, Number


class FetchedTile:
    """
    This is the contract for providing the data for constructing a :class:`slicedimage.Tile`.
    """
    def __init__(self, *args, **kwargs):
        pass

    @property
    def shape(self) -> Mapping[Axes, int]:
        """Return Tile shape.

        Returns
        -------
        Mapping[Axis, int]
            The shape of the tile, mapping from Axes to its size.
        """
        raise NotImplementedError()

    @property
    def coordinates(self) -> Mapping[Union[str, Coordinates], Union[Number, Tuple[Number, Number]]]:
        """Return the tile's coordinates in the global coordinate space..

        Returns
        -------
        Mapping[Union[str, Coordinates], Union[Number, Tuple[Number, Number]]]
            Maps from a coordinate type (e.g. 'x', 'y', or 'z') to its value or range.
        """
        raise NotImplementedError()

    @property
    def extras(self) -> dict:
        """Return the extras data associated with the tile.

        Returns
        -------
        Mapping[str, Any]
            Maps from a key to its value.
        """
        return {}

    def tile_data(self) -> np.ndarray:
        """Return the image data representing the tile.  The tile must be row-major.

        Returns
        -------
        ndarray :
            The image data
        """
        raise NotImplementedError()


class TileFetcher:
    """
    This is the contract for providing the image data for constructing a
    :class:`slicedimage.Collection`.
    """
    def get_tile(self, fov: int, r: int, ch: int, z: int) -> FetchedTile:
        """
        Given fov, r, ch, and z, return an instance of a :class:`.FetchedImage` that can be
        queried for the image data.
        """
        raise NotImplementedError()
